;; ARM code analyser
;; Copyright (C) 2006 Tomasz Malesinski <tmal@mimuw.edu.pl>

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

;; Some of the macros were taken from "On Lisp" book by Paul Graham

(defmacro aif (test-form then-form &optional else-form)
  `(let ((it ,test-form))
     (if it ,then-form ,else-form)))

(defmacro our-with-gensyms (syms &body body)
  `(let ,(mapcar #'(lambda (s)
		     `(,s (gensym)))
		 syms)
     ,@body))

(defmacro dbind (pat seq &body body)
  (let ((gseq (gensym)))
    `(let ((,gseq ,seq))
       ,(dbind-ex (destruc pat gseq #'atom) body))))

(defun destruc (pat seq &optional (atom? #'atom) (n 0))
  (if (null pat)
      nil
    (let ((rest (cond ((funcall atom? pat) pat)
		      ((eq (car pat) '&rest) (cadr pat))
		      ((eq (car pat) '&body) (cadr pat))
		      (t nil))))
      (if rest
	  `((,rest (subseq ,seq ,n)))
	(let ((p (car pat))
	      (rec (destruc (cdr pat) seq atom? (1+ n))))
	  (if (funcall atom? p)
	      (cons `(,p (elt ,seq ,n))
		    rec)
	    (let ((var (gensym)))
	      (cons (cons `(,var (elt ,seq ,n))
			  (destruc p var atom?))
		    rec))))))))

(defun vars-in (expr &optional (atom? #'atom))
  (if (funcall atom? expr)
      (if (var? expr) (list expr))
    (union (vars-in (car expr) atom?)
	   (vars-in (cdr expr) atom?))))

(defun var? (x)
  (and (symbolp x) (eq (char (symbol-name x) 0) #\?)))

(defun dbind-ex (binds body)
  (if (null binds)
      `(progn ,@body)
    `(let ,(mapcar #'(lambda (b)
		       (if (consp (car b))
			   (car b) b))
		   binds)
       ,(dbind-ex (mapcan #'(lambda (b)
			      (if (consp (car b))
				  (cdr b)))
			  binds)
		  body))))

(defmacro if-match (pat seq then &optional else)
  `(let ,(mapcar #'(lambda (v) `(,v ',(gensym)))
		 (vars-in pat #'simple?))
     (pat-match ,pat ,seq ,then ,else)))

(defmacro pat-match (pat seq then else)
  (if (simple? pat)
      (match1 `((,pat ,seq)) then else)
    (our-with-gensyms (gseq gelse)
		      `(labels ((,gelse () ,else))
			 ,(gen-match (cons (list gseq seq)
					   (destruc pat gseq #'simple?))
				     then
				     `(,gelse))))))

(defun simple? (x) (or (atom x) (eq (car x) 'quote)))

(defun gen-match (refs then else)
  (if (null refs)
      then
    (let ((then (gen-match (cdr refs) then else)))
      (if (simple? (caar refs))
	  (match1 refs then else)
	(gen-match (car refs) then else)))))

(defun match1 (refs then else)
  (dbind ((pat expr) . rest) refs
	 (cond ((gensym? pat)
		`(let ((,pat ,expr))
		   (if (and (typep ,pat 'sequence)
			    ,(length-test pat rest))
		       ,then
		     ,else)))
	       ((eq pat '_) then)
	       ((var? pat)
		(let ((ge (gensym)))
		  `(let ((,ge ,expr))
		     (if (or (gensym? ,pat) (equal ,pat ,ge))
			 (let ((,pat ,ge)) ,then)
		       ,else))))
	       (t `(if (equal ,pat ,expr) ,then ,else)))))

(defun gensym? (s)
  (and (symbolp s) (not (symbol-package s))))

(defun length-test (pat rest)
  (let ((fin (caadar (last rest))))
    (if (or (consp fin) (eq fin 'elt))
	`(= (length ,pat) ,(length rest))
      `(> (length ,pat) ,(- (length rest) 2)))))

(defmacro cond-matchg (expr &rest clauses)
  (let ((gexpr (gensym)))
    `(let ((,gexpr ,expr))
       ,(cond-matchg-ex-clauses clauses gexpr))))

(defun cond-matchg-ex-clauses (clauses gexpr)
  (if (null clauses)
      nil
    (let ((else (cond-matchg-ex-clauses (cdr clauses) gexpr)))
      (destructuring-bind ((pat &optional (guard t)) &body body)
	  (car clauses)
      (if (eq guard t)
	  `(if-match ,pat ,gexpr
		     (progn ,@body)
		     ,else)
	(let ((gelse (gensym)))
	  `(labels ((,gelse () ,else))
	     (if-match ,pat ,gexpr
		       (if ,guard
			   (progn ,@body)
			 (,gelse))
		       (,gelse)))))))))

(defmacro cond-match (expr &rest clauses)
  `(cond-matchg ,expr
		,@(mapcar #'(lambda (x) (cons (list (car x))
					      (cdr x)))
			  clauses)))

(defmacro match (pat seq &key (where t))
  `(if-match ,pat ,seq ,where nil))

(defmacro if-some-match (pat seq then &optional else)
  (let ((garg (gensym))
	(grec (gensym)))
    `(labels ((,grec (,garg)
		     (if (null ,garg)
			 ,else
		       (if-match ,pat (car ,garg)
				 (let ((it (car ,garg)))
				   ,then)
				 (,grec (cdr ,garg))))))
	     (,grec ,seq))))
